;; CORDIC
;; 278 Bytes
;; Ticks < 560 (inclusive call+ret)
;; Georg-Johann Lay

.include "cordic.inc"

__zero_reg__ = 1

.text

.global cordic
.type   cordic, @function

;; CORDIC-Algorithmus fr cos() und sin()
;; Register-Verwendung ist avr-gcc ABI-konform
_phi    = 26        ;;  Winkel:     16 Bit signed
_cos    = 22        ;;  cos:        16 Bit signed     R16..R22
_sin    = 24        ;;  sin:        16 Bit signed
_pow2   = 20        ;;  2^{-i}:      8 Bit unsigned   R16..R23
_zero   = 21        ;;  immer 0:     8 Bit
_tmp    = 18        ;; 16 Bit                         R16..R22
; R31:R30           ;; 16 Bit        cordic_atan_tab[]

/*
    cos = 1/1.6468 = 1 / CORDIC_GAIN
    sin = 0
    for i=0..15
        tmp = sin
        if (phi > 0)
            sin += cos * 2^(-i)
            cos -= tmp * 2^(-i)
            phi -= arctan(2^(-i))
        else
            sin -= cos * 2^(-i)
            cos += tmp * 2^(-i)
            phi += arctan(2^(-i))
        endif
    next i
*/

cordic:
    ;; end prolog
    movw    _phi, R24
    clr     _zero
    ;; Z = cordic_atan_tab[]
    ldi     r30, lo8 (cordic_atan_tab)
    ldi     r31, hi8 (cordic_atan_tab)
    ;;  sin = 0
    clr     _sin
    clr     _sin+1
    ;;  cos = 1 / CORDIC_GAIN
    lpm     _cos,   z+
    lpm     _cos+1, z+
    
    ldi     _pow2, 0x80
    
    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    ;; Shift = 0
    movw    _tmp, _sin;
    
    tst     _phi+1
    brmi    0f
    
    ;; phi > 0
    add         _sin,   _cos
    adc         _sin+1, _cos+1
    sub         _cos,   _tmp
    sbc         _cos+1, _tmp+1
    rjmp        .L_phi_pos_1_7
    
    ;; phi < 0
0:  sub         _sin,   _cos
    sbc         _sin+1, _cos+1
    add         _cos,   _tmp
    adc         _cos+1, _tmp+1
    rjmp        .L_phi_neg_1_7

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Shifts 1..7
.L_loop_1_7:
    movw    _tmp, _sin;
    
    tst     _phi+1
    brmi    0f
  
    ;; phi > 0
    MADD16_8    _sin  _cos  _pow2 _zero
    MSUB16_8    _cos  _tmp  _pow2 _zero
.L_phi_pos_1_7:
    PHI_SUB
    lsr     _pow2
    brne    .L_loop_1_7
    breq    .L_done_1_7

    ;; phi < 0
0:  MSUB16_8    _sin  _cos  _pow2 _zero
    MADD16_8    _cos  _tmp  _pow2 _zero
.L_phi_neg_1_7:
    ;; phi += atan (2^{-i})
    PHI_ADD
    lsr     _pow2
    brne    .L_loop_1_7

.L_done_1_7:
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Shifts 8..15
    ldi     _pow2, 0x80
    
.L_loop_8_15:
    movw    _tmp, _sin;
    
    tst     _phi+1
    brmi    0f

    ;; phi > 0
    MADD16_16   _sin  _cos  _pow2 _zero
    MSUB16_16   _cos  _tmp  _pow2 _zero
    PHI_SUB
    lsr     _pow2
    brne    .L_loop_8_15
    breq    .L_done_8_15

    ;; phi < 0
0:  MSUB16_16   _sin  _cos  _pow2 _zero
    MADD16_16   _cos  _tmp  _pow2 _zero
    PHI_ADD
    lsr     _pow2
    brne    .L_loop_8_15

.L_done_8_15:
    ;; avr-gcc ABI
    clr     __zero_reg__
    ;ret
    
    ;; Mittleren Fehler minimieren
    ;; sin += 1
.global cordic_round
cordic_round:    
    ;adiw    _sin, 1
    ;brvc    0f
    ;sbiw    _sin, 1
    ;; cos += 1
0:  subi    _cos,   lo8 (-1)
    sbci    _cos+1, hi8 (-1) 
    brvc    0f
    ldi     _cos,   lo8 (0x7fff)
    ldi     _cos+1, hi8 (0x7fff) 
0:  ret

.size	cordic, .-cordic
