#ifndef CORDIC_H
#define CORDIC_H

#include <stdint.h>

typedef struct
{
    int16_t cos;
    int16_t sin;
} cos_sin_t;

// Input in (-sqrt3, sqrt3) ~ (-99,99)
#define CORDIC_SCALE_IN(X)   (18000.*(X)+0.5)

// Output in (-1,1) als 1.15 signed Q-Format
#define CORDIC_SCALE_OUT(X)  (32767.*(X))

// Gain = \prod_i \sqrt{1+4^{-i}}
// 1 / Gain
#define CORDIC_GAIN 0.60725294

// -sqrt3 < phi < sqrt3
// Ticks < 560
// Standardabweichung:
// sigma (sin()) ~ 3   * 2^{-15}
// sigma (cos()) ~ 2.5 * 2^{-15}
extern cos_sin_t cordic (int16_t phi);

#endif // CORDIC_H
